/**
 * Copyright (C) 2010-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.gogrid.network;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.compute.VirtualMachine;
import org.dasein.cloud.jclouds.gogrid.GoGrid;
import org.dasein.cloud.network.AddressType;
import org.dasein.cloud.network.IpAddress;
import org.dasein.cloud.network.IpAddressSupport;
import org.dasein.cloud.network.IpForwardingRule;
import org.dasein.cloud.network.LoadBalancer;
import org.dasein.cloud.network.Protocol;
import org.dasein.util.CalendarWrapper;
import org.jclouds.gogrid.GoGridAsyncClient;
import org.jclouds.gogrid.GoGridClient;
import org.jclouds.gogrid.domain.Ip;
import org.jclouds.gogrid.domain.IpState;
import org.jclouds.gogrid.domain.IpType;
import org.jclouds.gogrid.options.GetIpListOptions;
import org.jclouds.rest.RestContext;

public class GoGridIp implements IpAddressSupport {
    static private final Logger logger = Logger.getLogger(GoGridIp.class);
    
    private GoGrid provider;
    
    GoGridIp(GoGrid gogrid) {
        provider = gogrid;
    }
    
    @Override
    public void assign(String addressId, String toServerId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("GoGrid does not support IP assignment.");        
    }

    @Override
    public String forward(String ipAddressId, int publicPort, Protocol protocol, int privatePort, String onServerId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("GoGrid does not support IP forwarding.");
    }

    static private final HashMap<String,Map<String,IpAddress>> publicIpCache = new HashMap<String,Map<String,IpAddress>>();
    static private final HashMap<String,Map<String,IpAddress>> privateIpCache = new HashMap<String,Map<String,IpAddress>>();
    static private final HashMap<String,Long> publicCacheTime = new HashMap<String,Long>();
    static private final HashMap<String,Long> privateCacheTime = new HashMap<String,Long>();
    
    private void cachePublicAddress(IpAddress addr) {
        Map<String,IpAddress> cache = publicIpCache.get(provider.getContext().getAccountNumber());
        
        if( cache == null ) {
            cache = new HashMap<String,IpAddress>();
            publicIpCache.put(provider.getContext().getAccountNumber(), cache);
        }
        cache.put(addr.getProviderIpAddressId(), addr);
    }
    
    private void cachePrivateAddresses(Collection<IpAddress> addresses) {
        HashMap<String,IpAddress> cache = new HashMap<String,IpAddress>();
        String acct = provider.getContext().getAccountNumber();
        
        for( IpAddress address : addresses ) {
            cache.put(address.getProviderIpAddressId(), address);
        }
        privateIpCache.put(acct, cache);
        privateCacheTime.put(acct, System.currentTimeMillis());
    }
    
    private void cachePublicAddresses(Collection<IpAddress> addresses) {
        HashMap<String,IpAddress> cache = new HashMap<String,IpAddress>();
        String acct = provider.getContext().getAccountNumber();
        
        for( IpAddress address : addresses ) {
            cache.put(address.getProviderIpAddressId(), address);
        }
        publicIpCache.put(acct, cache);
        publicCacheTime.put(acct, System.currentTimeMillis());
    }
    
    private IpAddress readAddressFromCache(String id) {
        Map<String,IpAddress> cache = publicIpCache.get(provider.getContext().getAccountNumber());
        
        if( cache == null ) {
            return null;
        }
        return cache.get(id);
    }
    
    private ArrayList<IpAddress> readPrivateFromCache() {
        Map<String,IpAddress> cache = privateIpCache.get(provider.getContext().getAccountNumber());
        Long when = privateCacheTime.get(provider.getContext().getAccountNumber());
        
        if( cache == null || when == null || (when.longValue() < (System.currentTimeMillis() - (CalendarWrapper.MINUTE*10L))) ) {
            privateIpCache.put(provider.getContext().getAccountNumber(), null);
            return new ArrayList<IpAddress>();
        }
        ArrayList<IpAddress> list = new ArrayList<IpAddress>();
        
        list.addAll(cache.values());
        return list;
    }
    
    private ArrayList<IpAddress> readPublicFromCache() {
        Map<String,IpAddress> cache = publicIpCache.get(provider.getContext().getAccountNumber());
        Long when = publicCacheTime.get(provider.getContext().getAccountNumber());
        
        if( cache == null || when == null || (when.longValue() < (System.currentTimeMillis() - (CalendarWrapper.MINUTE*10L))) ) {
            publicIpCache.put(provider.getContext().getAccountNumber(), null);
            return new ArrayList<IpAddress>();
        }
        ArrayList<IpAddress> list = new ArrayList<IpAddress>();
        
        list.addAll(cache.values());
        return list;
    }
    
    @Override
    public IpAddress getIpAddress(String ipAddressId) throws InternalException, CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
        IpAddress address = readAddressFromCache(ipAddressId);

        if( address != null ) {
            return address;
        }
        long id;
        
        try {
            id = Long.parseLong(ipAddressId);
        }
        catch( NumberFormatException e ) {
            return null;
        }
        try {
            count("getIpList");
            for( Ip ip : ctx.getApi().getIpServices().getIpList(GetIpListOptions.NONE) ) {
                if( ip.getId() == id ) {
                    address = toIpAddress(ip);
                    
                    if( address != null ) {
                        cachePublicAddress(address);
                    }
                    return address;
                }
            }
        }
        catch( RuntimeException e ) {
            logger.error("Could not find IP address " + ipAddressId + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Could not find IP address " + ipAddressId + ": " + e.getMessage());
        }
        finally {
            ctx.close();
        }
        return null;
    }

    @Override
    public String getProviderTermForIpAddress(Locale locale) {
        return "IP address";
    }

    @Override
    public boolean isAssigned(AddressType type) {
        return false;
    }

    @Override
    public boolean isForwarding() {
        return false;
    }

    @Override
    public boolean isRequestable(AddressType type) {
        return false;
    }
    
    @Override
    public boolean isSubscribed() {
        return true;
    }
    
    @Override 
    public Iterable<IpAddress> listPrivateIpPool(boolean unassignedOnly) throws InternalException, CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
        ArrayList<IpAddress> addresses = readPrivateFromCache();
        
        if( !addresses.isEmpty() ) {
            return addresses;
        }
        try {
            GetIpListOptions options = new GetIpListOptions();
            
            options.inDatacenter(provider.getContext().getRegionId());
            if( unassignedOnly ) {
                options.onlyUnassigned();
            }
            options.onlyWithType(IpType.PRIVATE);
            count("getIpList");
            for( Ip ip : ctx.getApi().getIpServices().getIpList(options) ) {
                if( ip != null && !ip.isPublic() ) {
                    IpAddress address = toIpAddress(ip);
                    
                    if( address != null ) {
                        addresses.add(address);
                    }
                }
            }
            cachePrivateAddresses(addresses);
        }
        catch( RuntimeException e ) {
            logger.error("Failed to list IP addresses: " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to list IP addresses: " + e.getMessage());
        }
        finally {
            ctx.close();
        }
        return addresses;
    }
    
    @Override
    public Collection<IpAddress> listPublicIpPool(boolean unassignedOnly) throws InternalException, CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
        ArrayList<IpAddress> addresses = readPublicFromCache();
        
        if( !addresses.isEmpty() ) {
            return addresses;
        }
        try {
            GetIpListOptions options = new GetIpListOptions();
            IpAddress highest = null;
            
            options.inDatacenter(provider.getContext().getRegionId());
            if( unassignedOnly ) {
                options.onlyUnassigned();
            }
            options.onlyWithType(IpType.PUBLIC);
            count("getIpList");
            for( Ip ip : ctx.getApi().getIpServices().getIpList(options) ) {
                if( ip != null && ip.isPublic() ) {
                    IpAddress address = toIpAddress(ip);
                    
                    if( address != null ) {
                        if( highest != null ) {
                            if( highest.getAddress().compareTo(address.getAddress()) > 0 ) {
                                addresses.add(address);
                            }
                            else {
                                addresses.add(highest);
                                highest = address;
                            }
                        }
                        highest = address;
                    }
                }
            }
            cachePublicAddresses(addresses);
        }
        catch( RuntimeException e ) {
            logger.error("Failed to list IP addresses: " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to list IP addresses: " + e.getMessage());
        }
        finally {
            ctx.close();
        }
        return addresses;
    }

    @Override
    public Collection<IpForwardingRule> listRules(String ipAddressId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("GoGrid does not support IP forwarding.");
    }

    @Override
    public void releaseFromPool(String ipAddressId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("GoGrid does not support IP assignment.");
    }
    
    @Override
    public void releaseFromServer(String ipAddressId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("GoGrid does not support manual assignment.");
    }

    @Override
    public String request(AddressType type) throws InternalException, CloudException {
        throw new OperationNotSupportedException("GoGrid does not support IP allocation.");        
    }
    
    @Override
    public void stopForward(String ipAddressId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("GoGrid uses IP address assignment.");
    }

    private IpAddress toIpAddress(Ip ip) throws CloudException, InternalException {
        IpAddress address = new IpAddress();

        address.setAddress(ip.getIp());
        address.setIpAddressId(String.valueOf(ip.getId()));
        address.setAddressType(ip.isPublic() ? AddressType.PUBLIC : AddressType.PRIVATE);
        address.setRegionId(provider.getContext().getRegionId());
        address.setServerId(null);
        if( ip.getState().equals(IpState.ASSIGNED) ) {
            for( VirtualMachine svr : provider.getComputeServices().getVirtualMachineSupport().listVirtualMachines() ) {
                String id = svr.getProviderAssignedIpAddressId();

                if( id != null && Long.parseLong(id) == ip.getId() ) {
                    address.setServerId(svr.getProviderVirtualMachineId());
                }
            }
        }
        if( ip.isPublic() ) {
            for( LoadBalancer lb : provider.getNetworkServices().getLoadBalancerSupport().listLoadBalancers() ) {
                if( lb.getAddress().equals(ip.getIp()) ) {
                    address.setProviderLoadBalancerId(lb.getProviderLoadBalancerId());
                    break;
                }
            }
        }
        return address;
    }
    
    static private final Logger counter = Logger.getLogger("gogrid.count");
    
    private void count(String method) {
        counter.info(method);
    }
}
