/**
 * Copyright (C) 2010-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.gogrid.network;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.concurrent.TimeUnit;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.compute.VirtualMachine;
import org.dasein.cloud.jclouds.gogrid.GoGrid;
import org.dasein.cloud.network.IpAddress;
import org.dasein.cloud.network.LbAlgorithm;
import org.dasein.cloud.network.LbListener;
import org.dasein.cloud.network.LbProtocol;
import org.dasein.cloud.network.LoadBalancer;
import org.dasein.cloud.network.LoadBalancerAddressType;
import org.dasein.cloud.network.LoadBalancerSupport;
import org.jclouds.gogrid.GoGridAsyncClient;
import org.jclouds.gogrid.GoGridClient;
import org.jclouds.gogrid.GoGridResponseException;
import org.jclouds.gogrid.domain.Ip;
import org.jclouds.gogrid.domain.IpPortPair;
import org.jclouds.gogrid.domain.Job;
import org.jclouds.gogrid.domain.JobState;
import org.jclouds.gogrid.domain.LoadBalancerPersistenceType;
import org.jclouds.gogrid.domain.LoadBalancerType;
import org.jclouds.gogrid.domain.ObjectType;
import org.jclouds.gogrid.options.AddLoadBalancerOptions;
import org.jclouds.gogrid.options.GetJobListOptions;
import org.jclouds.gogrid.options.GetServerListOptions;
import org.jclouds.gogrid.predicates.LoadBalancerLatestJobCompleted;
import org.jclouds.predicates.RetryablePredicate;
import org.jclouds.rest.ResourceNotFoundException;
import org.jclouds.rest.RestContext;

public class GoGridLoadBalancer implements LoadBalancerSupport {
    static private final Logger logger = Logger.getLogger(GoGridLoadBalancer.class);
    
    private final GoGrid provider;
    
    GoGridLoadBalancer(GoGrid gogrid) {
        provider = gogrid;
    }
    
    @Override
    public void addDataCenters(String loadBalancerId, String ... dataCenterIds) throws CloudException, InternalException {
        // NO-OP
    }

    @Override
    public void addServers(String loadBalancerId, String ... serverIds) throws CloudException, InternalException {
        if( serverIds == null || serverIds.length < 1 ) {
            return;
        }
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();

        try {   
            waitForLoadBalancer(ctx, loadBalancerId);
            
            org.jclouds.gogrid.domain.LoadBalancer loadBalancer = getGoGridLoadBalancer(ctx, Long.parseLong(loadBalancerId));
            LoadBalancer lb = toLoadBalancer(ctx, loadBalancer);
            ArrayList<IpPortPair> newIps = new ArrayList<IpPortPair>();
            Set<IpPortPair> current = loadBalancer.getRealIpList();
            LbListener[] listeners = lb.getListeners();
            
            for( IpPortPair pair : current ) {
                newIps.add(pair);
            }
            for( String serverId : serverIds ) {
                Ip ip = getIp(serverId);
                
                if( ip == null ) {
                    continue;
                }
                newIps.add(new IpPortPair(ip, listeners[0].getPrivatePort()));
            }
            count("editLoadBalancer");
            loadBalancer = ctx.getApi().getLoadBalancerServices().editLoadBalancer(Long.parseLong(loadBalancerId), newIps);
            
            count("getJobServices");
            RetryablePredicate<org.jclouds.gogrid.domain.LoadBalancer> wait = new RetryablePredicate<org.jclouds.gogrid.domain.LoadBalancer>(new LoadBalancerLatestJobCompleted(ctx.getApi().getJobServices()), 800, 20, TimeUnit.SECONDS);
            
            if( wait.apply(loadBalancer) ) {
                return;
            }
            throw new CloudException("Load balancer changes failed.");
        }
        catch( GoGridResponseException e ) {
            logger.error("Failed to delete load balancer " + loadBalancerId + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException(e);
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public String create(String name, String description, String addressId, String[] dcIds, LbListener[] listeners, String[] servers) throws CloudException, InternalException {
        try {
            IpAddress address = provider.getNetworkServices().getIpAddressSupport().getIpAddress(addressId);

            AddLoadBalancerOptions options = new AddLoadBalancerOptions();
            List<IpPortPair> mappings = new ArrayList<IpPortPair>();
            IpPortPair pub;

            if( listeners == null || listeners.length != 1 ) {
                throw new CloudException("You can specify only one listener for GoGrid.");
            }
            if( address == null ) {
                throw new CloudException("No such address: " + addressId);
            }
            Ip ip = new Ip(address.getAddress());
            pub = new IpPortPair(ip, listeners[0].getPublicPort());
            options.setPersistenceType(LoadBalancerPersistenceType.NONE);
            options.setDescription(description);
            options.setType(getType(listeners[0].getAlgorithm()));
            for( String server : servers ) {
                ip = getIp(server);
                if( ip != null ) {
                    mappings.add(new IpPortPair(ip, listeners[0].getPrivatePort()));
                }
            }

            RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();

            try {
                try {
                    count("addLoadBalancer");
                    ctx.getApi().getLoadBalancerServices().addLoadBalancer(name, pub, mappings, options);
                }
                catch( GoGridResponseException e ) {
                    logger.error("Failed to add load balancer: " + e.getMessage());
                    if( logger.isDebugEnabled() ) {
                        e.printStackTrace();
                    }
                    throw new CloudException(e);
                }

                org.jclouds.gogrid.domain.LoadBalancer loadBalancer = null;
                
                while( loadBalancer == null ) {
                    Iterable<org.jclouds.gogrid.domain.LoadBalancer> loadBalancers;
                    
                    try {
                        count("getLoadBalancersByName");
                        loadBalancers = ctx.getApi().getLoadBalancerServices().getLoadBalancersByName(name);
                    }
                    catch( ResourceNotFoundException ignore ) {
                        continue;
                    }
                    for( org.jclouds.gogrid.domain.LoadBalancer lb : loadBalancers ) {
                        if( lb.getId() != 0L ) {
                            loadBalancer = lb;
                            break;
                        }
                    }
                    if( loadBalancer == null ) {
                        try { Thread.sleep(10000L); }
                        catch( InterruptedException e ) { }
                    }
                }
                return toLoadBalancer(ctx, loadBalancer).getProviderLoadBalancerId();
            }
            finally {
                ctx.close();
            }
        }
        catch( RuntimeException e ) {
            logger.error("Failed to add load balancer: " + e.getMessage());
            if( logger.isDebugEnabled() ) {
                e.printStackTrace();
            }
            throw new InternalException(e);
        }
        catch( Error e ) {
            logger.error("Failed to add load balancer: " + e.getMessage());
            if( logger.isDebugEnabled() ) {
                e.printStackTrace();
            }
            throw new InternalException(e);
        }
    }

    @Override
    public LoadBalancerAddressType getAddressType() {
        return LoadBalancerAddressType.IP;
    }
    
    private Ip getIp(String serverId) throws InternalException, CloudException {
        VirtualMachine server = provider.getComputeServices().getVirtualMachineSupport().getVirtualMachine(serverId);
        
        if( server == null ) {
            return null;
        }
        String[] ips = server.getPrivateIpAddresses();
        
        if( ips == null || ips.length < 1 ) {
            ips = server.getPublicIpAddresses();
            if( ips == null || ips.length < 1 ) {
                return null;
            }
        }
        
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
        
        try {
            count("getIpList");
            for( Ip ip : ctx.getApi().getIpServices().getIpList() ) {
                if( ip.getIp().equals(ips[0]) ) {
                    return ip;
                }
            }
            return null;
        }
        finally {
            ctx.close();
        }
    }
    
    @Override
    public LoadBalancer getLoadBalancer(String loadBalancerId) throws CloudException, InternalException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
        
        try {
            long id;
            
            try {
                id = Long.parseLong(loadBalancerId);
            }
            catch( NumberFormatException unimportant ) {
                return null;
            }
            return toLoadBalancer(ctx, getGoGridLoadBalancer(ctx, id));
        }
        catch( RuntimeException e ) {
            String msg = e.getMessage();
            
            if( msg != null && msg.contains("No object found") ) {
                return null;
            }
            logger.error("Could not list load balancers from GoGrid: " + e.getMessage());
            e.printStackTrace();
            throw new CloudException(e);
        }
        finally {
            ctx.close();
        }
    }
    
    private org.jclouds.gogrid.domain.LoadBalancer getGoGridLoadBalancer(RestContext<GoGridClient,GoGridAsyncClient> ctx, long loadBalancerId) throws CloudException, InternalException {
        Set<org.jclouds.gogrid.domain.LoadBalancer> loadBalancers;
            
        count("getLoadBalancersById");
        loadBalancers = ctx.getApi().getLoadBalancerServices().getLoadBalancersById(loadBalancerId);
        for( org.jclouds.gogrid.domain.LoadBalancer loadBalancer : loadBalancers ) {
            if( loadBalancer.getId() == loadBalancerId ) {
                return loadBalancer;
            }
        }
        return null;
    }

    @Override
    public int getMaxPublicPorts() throws CloudException, InternalException {
        return 1;
    }
    
    @Override
    public String getProviderTermForLoadBalancer(Locale locale) {
        return "load balancer";
    }

    private String getServerId(Ip ip) throws InternalException, CloudException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();
        
        try {
            count("getServerList");
            for( org.jclouds.gogrid.domain.Server fromCloud : ctx.getApi().getServerServices().getServerList(new GetServerListOptions()) ) {
                if( fromCloud.getIp() != null && fromCloud.getIp().getIp().equals(ip.getIp()) ) {
                    return String.valueOf(fromCloud.getId());
                }
            }
        }
        finally {
            ctx.close();
        }
        return null;
    }

    private boolean isBusy(RestContext<GoGridClient,GoGridAsyncClient> ctx, String loadBalancerId) throws InternalException, CloudException {
        long id = Long.parseLong(loadBalancerId);
        org.jclouds.gogrid.domain.LoadBalancer loadBalancer = getGoGridLoadBalancer(ctx, id);
        GetJobListOptions options = new GetJobListOptions();
        
        options.onlyForObjectType(ObjectType.LOAD_BALANCER);
        options.onlyForObjectName(loadBalancer.getName());
        count("getJobList");
        for( Job job : ctx.getApi().getJobServices().getJobList(options) ) {
            JobState s = job.getCurrentState();
            
            if( s.equals(JobState.CREATED) && s.equals(JobState.PROCESSING) && s.equals(JobState.QUEUED) && s.equals(JobState.UNRECOGNIZED) ) {
                return true;
            }
        }
        return false;
    }
    
    static private volatile List<LbAlgorithm> algorithms = null;
    
    public Iterable<LbAlgorithm> listSupportedAlgorithms() {
        List<LbAlgorithm> list = algorithms;
        
        if( list == null ) {
            list = new ArrayList<LbAlgorithm>();
            list.add(LbAlgorithm.ROUND_ROBIN);
            list.add(LbAlgorithm.LEAST_CONN);
            algorithms = Collections.unmodifiableList(list);
        }
        return algorithms;
    }
    
    static private volatile List<LbProtocol> protocols = null;
    
    public Iterable<LbProtocol> listSupportedProtocols() {
        List<LbProtocol> list = protocols;
        
        if( protocols == null ) {
            list = new ArrayList<LbProtocol>();
            list.add(LbProtocol.RAW_TCP);
            protocols = Collections.unmodifiableList(list);
        }
        return protocols;
    }
    
    private LoadBalancerType getType(LbAlgorithm algorithm) {
        switch( algorithm ) {
            case ROUND_ROBIN: return LoadBalancerType.ROUND_ROBIN;
            case LEAST_CONN: return LoadBalancerType.LEAST_CONNECTED;
            case SOURCE: return LoadBalancerType.UNRECOGNIZED;
        }
        return LoadBalancerType.UNRECOGNIZED;
    }
    
    @Override
    public boolean isAddressAssignedByProvider() throws CloudException, InternalException {
        return false;
    }

    @Override
    public boolean isDataCenterLimited() {
        return false;
    }
    
    @Override
    public boolean requiresListenerOnCreate() throws CloudException, InternalException {
        return true;
    }

    @Override
    public boolean requiresServerOnCreate() throws CloudException, InternalException {
        return true;
    }

    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
        return true;
    }

    @Override
    public boolean supportsMonitoring() throws CloudException, InternalException {
        return false;
    }

    @Override
    public Collection<LoadBalancer> listLoadBalancers() throws CloudException, InternalException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();

        try {   
            count("getLoadBalancerList");
            Set<org.jclouds.gogrid.domain.LoadBalancer> loadBalancers = ctx.getApi().getLoadBalancerServices().getLoadBalancerList();
            HashMap<String,LoadBalancer> building = new HashMap<String,LoadBalancer>();
            
            for( org.jclouds.gogrid.domain.LoadBalancer loadBalancer : loadBalancers ) {
                String addr = loadBalancer.getVirtualIp().getIp().getIp();
                LoadBalancer lb;
                
                if( building.containsKey(addr) ) {
                    lb = building.get(addr);
                    if( lb == null ) {
                        lb = toLoadBalancer(ctx, loadBalancer);
                    }
                    else {
                        lb = toListener(lb, loadBalancer);
                    }
                }
                else {
                    lb = toLoadBalancer(ctx, loadBalancer);
                }
                building.put(addr, lb);
            }
            return building.values();
        }
        catch( RuntimeException e ) {
            logger.error("Could not list load balancers from GoGrid: " + e.getMessage());
            e.printStackTrace();
            throw new CloudException(e);
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void remove(String loadBalancerId) throws CloudException, InternalException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();

        try {   
            waitForLoadBalancer(ctx, loadBalancerId);
            count("deleteById");
            ctx.getApi().getLoadBalancerServices().deleteById(Long.parseLong(loadBalancerId));
        }
        catch( RuntimeException e ) {
            logger.error("Failed to delete load balancer " + loadBalancerId + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException(e);
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void removeDataCenters(String loadBalancerId, String ... dcId) throws CloudException, InternalException {
        throw new OperationNotSupportedException("No support for removing data centers from a load balancer in GoGrid.");
    }

    @Override
    public void removeServers(String loadBalancerId, String ... serverIds) throws CloudException, InternalException {
        RestContext<GoGridClient,GoGridAsyncClient> ctx = provider.getCloudClient();

        try {
            waitForLoadBalancer(ctx, loadBalancerId);
            
            org.jclouds.gogrid.domain.LoadBalancer loadBalancer = getGoGridLoadBalancer(ctx, Long.parseLong(loadBalancerId));
            ArrayList<IpPortPair> newIps = new ArrayList<IpPortPair>();
            Set<IpPortPair> current = loadBalancer.getRealIpList();
            
            for( IpPortPair pair : current ) {
                String serverId = getServerId(pair.getIp());
                boolean keep = true;
                
                if( serverId != null ) {
                    for( String id : serverIds ) {
                        if( id.equals(serverId) ) {
                            keep = false;
                        }
                    }
                }
                if( keep ) {
                    newIps.add(pair);
                }
            }

            try {
                count("editLoadBalancer");
                loadBalancer = ctx.getApi().getLoadBalancerServices().editLoadBalancer(Long.parseLong(loadBalancerId), newIps);
            }
            catch( GoGridResponseException e ) {
                logger.warn("Failed to remove servers from load balancer " + loadBalancerId + ": " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException(e);
            }            
            count("getJobServices");
            RetryablePredicate<org.jclouds.gogrid.domain.LoadBalancer> wait = new RetryablePredicate<org.jclouds.gogrid.domain.LoadBalancer>(new LoadBalancerLatestJobCompleted(ctx.getApi().getJobServices()), 800, 20, TimeUnit.SECONDS);
            
            try {
                if( wait.apply(loadBalancer) ) {
                    return;
                }
            }
            catch( GoGridResponseException e ) {
                logger.warn("Failed to determine load balancer server removal status for " + loadBalancerId + ": " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException(e);                    
            }
            throw new CloudException("Load balancer changes failed.");

        }
        catch( RuntimeException e ) {
            logger.error("Failed remove server from load balancer: " + e.getMessage());
            if( logger.isDebugEnabled() ) {
                e.printStackTrace();
            }
            throw new InternalException(e);
        }
        catch( Error e ) {
            logger.error("Failed remove server from load balancer: " + e.getMessage());
            if( logger.isDebugEnabled() ) {
                e.printStackTrace();
            }
            throw new InternalException(e);
        }
        finally {
            ctx.close();
        }
    }

    private LbAlgorithm toAlgorithm(LoadBalancerType type) {
        switch( type ) {
            case ROUND_ROBIN: return LbAlgorithm.ROUND_ROBIN;
            case LEAST_CONNECTED: return LbAlgorithm.LEAST_CONN;
            case UNRECOGNIZED: return LbAlgorithm.ROUND_ROBIN;
        }
        return LbAlgorithm.ROUND_ROBIN;
    }
    
    private LoadBalancer toListener(LoadBalancer current, org.jclouds.gogrid.domain.LoadBalancer fromGoGrid) throws InternalException, CloudException {
        LbListener[] currentListeners = current.getListeners();
        LbListener[] listeners = new LbListener[currentListeners.length+1];
        LbListener listener = null;
        
        for( int i=0; i<currentListeners.length; i++ ) {
            listeners[i] = currentListeners[i];
        }
        for( IpPortPair mapping : fromGoGrid.getRealIpList() ) {
            listener = new LbListener();
            listener.setPublicPort(fromGoGrid.getVirtualIp().getPort());
            listener.setPrivatePort(mapping.getPort());
            listener.setNetworkProtocol(LbProtocol.RAW_TCP);
            listener.setAlgorithm(toAlgorithm(fromGoGrid.getType()));
        }
        if( listener != null ) {
            current.setListeners(currentListeners);
        }
        return current;
    }
    
    private LoadBalancer toLoadBalancer(RestContext<GoGridClient,GoGridAsyncClient> ctx, org.jclouds.gogrid.domain.LoadBalancer loadBalancer) throws InternalException, CloudException {
        if( loadBalancer == null ) {
            return null;
        }
        String id = String.valueOf(loadBalancer.getId());
        String description = loadBalancer.getDescription();
        String name = loadBalancer.getName();
        LoadBalancer lb = new LoadBalancer();
        
        if( name == null ) {
            name = id;
        }
        if( description == null ) {
            description = name;
        }
        lb.setProviderOwnerId(provider.getContext().getAccountNumber());
        lb.setAddressType(LoadBalancerAddressType.IP);
        lb.setCurrentState(org.dasein.cloud.network.LoadBalancerState.ACTIVE);
        lb.setAddress(loadBalancer.getVirtualIp().getIp().getIp());
        lb.setDescription(description);
        lb.setName(name);
        lb.setProviderLoadBalancerId(id);

        ArrayList<String> servers = new ArrayList<String>();
        LbListener listener = null;
        
        for( IpPortPair mapping : loadBalancer.getRealIpList() ) {
            String serverId = getServerId(mapping.getIp());
            
            if( serverId != null ) {
                servers.add(serverId);
            }
            if( listener == null ) {
                listener = new LbListener();
                listener.setPublicPort(loadBalancer.getVirtualIp().getPort());
                listener.setPrivatePort(mapping.getPort());
                listener.setNetworkProtocol(LbProtocol.RAW_TCP);
                listener.setAlgorithm(toAlgorithm(loadBalancer.getType()));
            }
            
        }
        if( listener != null ) {
            lb.setListeners(new LbListener[] { listener });
            lb.setPublicPorts(new int[] { listener.getPublicPort() });
        }
        else {
            lb.setListeners(new LbListener[0]);
            lb.setPublicPorts(new int[0]);
        }
        lb.setProviderServerIds(servers.toArray(new String[servers.size()]));
        
        lb.setProviderDataCenterIds(new String[]{ String.valueOf(loadBalancer.getDatacenter().getId()) + "a" });
        lb.setProviderRegionId(provider.getContext().getRegionId());
        GetJobListOptions options = new GetJobListOptions();
        long ts = System.currentTimeMillis();
        
        options.onlyForObjectType(ObjectType.LOAD_BALANCER);
        options.onlyForObjectName(loadBalancer.getName());
        count("getJobList");
        for( Job job : ctx.getApi().getJobServices().getJobList(options) ) {
            Date d = job.getCreatedOn();
            
            if( d != null && d.getTime() < ts ) {
                ts = d.getTime();
            }
        }
        lb.setCreationTimestamp(ts);
        return lb;
    }
    
    private void waitForLoadBalancer(RestContext<GoGridClient,GoGridAsyncClient> ctx, String id) throws InternalException, CloudException {
        while( isBusy(ctx, id) ) {
            try { Thread.sleep(1000L); }
            catch( InterruptedException e ) { }
        }
    }
    
    static private final Logger counter = Logger.getLogger("gogrid.count");
    
    private void count(String method) {
        counter.info(method);
    }
}
